"""Cloud mappings for offline use."""

# pylint: disable=line-too-long

cloud_mappings_offline = {
    # All endpoints were retrieved with api-version 2023-01-01
    "global": {
        "portal": "https://portal.azure.com",
        "authentication": {
            "loginEndpoint": "https://login.microsoftonline.com",
            "audiences": [
                "https://management.core.windows.net/",
                "https://management.azure.com/",
            ],
            "tenant": "common",
            "identityProvider": "AAD",
        },
        "media": "https://rest.media.azure.net",
        "graphAudience": "https://graph.windows.net/",
        "graph": "https://graph.windows.net/",
        "name": "AzureCloud",
        "suffixes": {
            "azureDataLakeStoreFileSystem": "azuredatalakestore.net",
            "acrLoginServer": "azurecr.io",
            "sqlServerHostname": "database.windows.net",
            "azureDataLakeAnalyticsCatalogAndJob": "azuredatalakeanalytics.net",
            "keyVaultDns": "vault.azure.net",
            "storage": "core.windows.net",
            "azureFrontDoorEndpointSuffix": "azurefd.net",
            "storageSyncEndpointSuffix": "afs.azure.net",
            "mhsmDns": "managedhsm.azure.net",
            "mysqlServerEndpoint": "mysql.database.azure.com",
            "postgresqlServerEndpoint": "postgres.database.azure.com",
            "mariadbServerEndpoint": "mariadb.database.azure.com",
            "synapseAnalytics": "dev.azuresynapse.net",
            "attestationEndpoint": "attest.azure.net",
        },
        "batch": "https://batch.core.windows.net/",
        "resourceManager": "https://management.azure.com/",
        "vmImageAliasDoc": "https://raw.githubusercontent.com/Azure/azure-rest-api-specs/master/arm-compute/quickstart-templates/aliases.json",
        "activeDirectoryDataLake": "https://datalake.azure.net/",
        "sqlManagement": "https://management.core.windows.net:8443/",
        "microsoftGraphResourceId": "https://graph.microsoft.com/",
        "appInsightsResourceId": "https://api.applicationinsights.io",
        "appInsightsTelemetryChannelResourceId": (
            "https://dc.applicationinsights.azure.com/v2/track"
        ),
        "attestationResourceId": "https://attest.azure.net",
        "synapseAnalyticsResourceId": "https://dev.azuresynapse.net",
        "logAnalyticsResourceId": "https://api.loganalytics.io",
        "ossrDbmsResourceId": "https://ossrdbms-aad.database.windows.net",
    },
    "usgov": {
        "portal": "https://portal.azure.us",
        "authentication": {
            "loginEndpoint": "https://login.microsoftonline.us",
            "audiences": [
                "https://management.core.usgovcloudapi.net",
                "https://management.usgovcloudapi.net",
            ],
            "tenant": "common",
            "identityProvider": "AAD",
        },
        "media": "https://rest.media.usgovcloudapi.net",
        "graphAudience": "https://graph.windows.net/",
        "graph": "https://graph.windows.net/",
        "name": "AzureUSGovernment",
        "suffixes": {
            "acrLoginServer": "azurecr.us",
            "sqlServerHostname": "database.usgovcloudapi.net",
            "keyVaultDns": "vault.usgovcloudapi.net",
            "storage": "core.usgovcloudapi.net",
            "storageSyncEndpointSuffix": "afs.azure.us",
            "mhsmDns": "managedhsm.usgovcloudapi.net",
            "mysqlServerEndpoint": "mysql.database.usgovcloudapi.net",
            "postgresqlServerEndpoint": "postgres.database.usgovcloudapi.net",
            "mariadbServerEndpoint": "mariadb.database.usgovcloudapi.net",
            "synapseAnalytics": "dev.azuresynapse.usgovcloudapi.net",
        },
        "batch": "https://batch.core.usgovcloudapi.net",
        "resourceManager": "https://management.usgovcloudapi.net",
        "vmImageAliasDoc": "https://raw.githubusercontent.com/Azure/azure-rest-api-specs/master/arm-compute/quickstart-templates/aliases.json",
        "sqlManagement": "https://management.core.usgovcloudapi.net:8443",
        "microsoftGraphResourceId": "https://graph.microsoft.us/",
        "appInsightsResourceId": "https://api.applicationinsights.us",
        "appInsightsTelemetryChannelResourceId": (
            "https://dc.applicationinsights.us/v2/track"
        ),
        "synapseAnalyticsResourceId": "https://dev.azuresynapse.usgovcloudapi.net",
        "logAnalyticsResourceId": "https://api.loganalytics.us",
        "ossrDbmsResourceId": "https://ossrdbms-aad.database.usgovcloudapi.net",
    },
    "cn": {
        "portal": "https://portal.azure.cn",
        "authentication": {
            "loginEndpoint": "https://login.chinacloudapi.cn",
            "audiences": [
                "https://management.core.chinacloudapi.cn",
                "https://management.chinacloudapi.cn",
            ],
            "tenant": "common",
            "identityProvider": "AAD",
        },
        "media": "https://rest.media.chinacloudapi.cn",
        "graphAudience": "https://graph.chinacloudapi.cn/",
        "graph": "https://graph.chinacloudapi.cn/",
        "name": "AzureChinaCloud",
        "suffixes": {
            "acrLoginServer": "azurecr.cn",
            "sqlServerHostname": "database.chinacloudapi.cn",
            "keyVaultDns": "vault.azure.cn",
            "storage": "core.chinacloudapi.cn",
            "mhsmDns": "managedhsm.azure.cn",
            "mysqlServerEndpoint": "mysql.database.chinacloudapi.cn",
            "postgresqlServerEndpoint": "postgres.database.chinacloudapi.cn",
            "synapseAnalytics": "dev.azuresynapse.azure.cn",
        },
        "batch": "https://batch.chinacloudapi.cn",
        "resourceManager": "https://management.chinacloudapi.cn",
        "vmImageAliasDoc": "https://raw.githubusercontent.com/Azure/azure-rest-api-specs/master/arm-compute/quickstart-templates/aliases.json",
        "sqlManagement": "https://management.core.chinacloudapi.cn:8443",
        "microsoftGraphResourceId": "https://microsoftgraph.chinacloudapi.cn",
        "appInsightsResourceId": "https://api.applicationinsights.azure.cn",
        "appInsightsTelemetryChannelResourceId": (
            "https://dc.applicationinsights.azure.cn/v2/track"
        ),
        "synapseAnalyticsResourceId": "https://dev.azuresynapse.azure.cn",
        "logAnalyticsResourceId": "https://api.loganalytics.azure.cn",
        "ossrDbmsResourceId": "https://ossrdbms-aad.database.chinacloudapi.cn",
    },
}
