# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Deprecated location for TI Providers."""

from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"
